package ru.dbotthepony.kstarbound.defs.item

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import org.classdump.luna.ByteString
import org.classdump.luna.LuaRuntimeException
import org.classdump.luna.Table
import org.classdump.luna.TableFactory
import ru.dbotthepony.kstarbound.lua.StateMachine
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.toJsonObject
import ru.dbotthepony.kstarbound.util.KOptional
import ru.dbotthepony.kstarbound.json.get
import java.util.function.Supplier

fun ItemDescriptor(data: JsonElement): ItemDescriptor {
	if (data is JsonPrimitive) {
		return ItemDescriptor(data.asString, 1L)
	} else if (data is JsonArray) {
		val name = data[0].asString
		val count = data.get(1, 1L)
		val parameters = data.get(2, ::JsonObject)
		return ItemDescriptor(name, count, parameters)
	} else if (data is JsonObject) {
		val name = (data.get("name") ?: data.get("item") ?: throw JsonSyntaxException("Missing item name")).asString
		val count = data.get("count", 1L)
		val parameters = data.get("parameters") { data.get("parameters", ::JsonObject) }
		return ItemDescriptor(name, count, parameters)
	} else if (data is JsonNull) {
		return ItemDescriptor("air", 0L)
	} else {
		throw JsonSyntaxException("Invalid item descriptor: $data")
	}
}

fun ItemDescriptor(data: Table, stateMachine: StateMachine): Supplier<ItemDescriptor> {
	val name = stateMachine.index(data, 1L, "name", "item")
	val count = stateMachine.optionalIndex(data, 2L, "count")
	val parameters = stateMachine.optionalIndex(data, 3L, "parameters", "data")
	var result: KOptional<ItemDescriptor> = KOptional.empty()

	stateMachine.add {
		val iname = name.get()
		val icount = count.get().orElse(1L)
		val iparameters = parameters.get().map { if (it is Table) it.toJsonObject() else throw LuaRuntimeException("Invalid item descriptor parameters ($it)") }.orElse { JsonObject() }

		if (iname !is ByteString) throw LuaRuntimeException("Invalid item descriptor name (${iname})")
		if (icount !is Number) throw LuaRuntimeException("Invalid item descriptor count (${icount})")

		result = KOptional(ItemDescriptor(iname.decode(), icount.toLong(), iparameters))
	}

	return Supplier { result.value }
}

data class ItemDescriptor(
	val name: String,
	val count: Long,
	val parameters: JsonObject = JsonObject()
) {
	val isEmpty get() = count <= 0L

	fun toJson(): JsonObject? {
		if (isEmpty) {
			return null
		}

		return JsonObject().also {
			it.add("name", JsonPrimitive(name))
			it.add("count", JsonPrimitive(count))
			it.add("parameters", parameters.deepCopy())
		}
	}

	fun toTable(allocator: TableFactory): Table? {
		if (isEmpty) {
			return null
		}

		return allocator.newTable(0, 3).also {
			it.rawset("name", name)
			it.rawset("count", count)
			it.rawset("parameters", allocator.from(parameters))
		}
	}
}
